@@ -7,12 +7,14 @@
 // modified, or distributed except according to those terms.
 
 use futures_util::FutureExt;
+use keyed_priority_queue::KeyedPriorityQueue;
 use tokio::sync::mpsc;
 
 use std::{
+    borrow::Borrow,
+    cmp::Reverse,
     collections::VecDeque,
-    convert::TryFrom,
-    pin::Pin,
+    hash::{Hash, Hasher},
     str::FromStr,
     sync::{atomic, Arc, Mutex},
     task::{Context, Poll, Waker},
@@ -23,7 +25,7 @@ use crate::{
     conn::{pool::futures::*, Conn},
     error::*,
     opts::{Opts, PoolOpts},
-    queryable::transaction::{Transaction, TxOpts, TxStatus},
+    queryable::transaction::{Transaction, TxOpts},
 };
 
 mod recycler;
@@ -41,6 +43,15 @@ struct IdlingConn {
 }
 
 impl IdlingConn {
+    /// Returns true when this connection has a TTL and it elapsed.
+    fn expired(&self) -> bool {
+        self.conn
+            .inner
+            .ttl_deadline
+            .map(|t| Instant::now() > t)
+            .unwrap_or_default()
+    }
+
     /// Returns duration elapsed since this connection is idling.
     fn elapsed(&self) -> Duration {
         self.since.elapsed()
@@ -62,7 +73,7 @@ impl From<Conn> for IdlingConn {
 /// This is fine as long as we never do expensive work while holding the lock!
 #[derive(Debug)]
 struct Exchange {
-    waiting: VecDeque<Waker>,
+    waiting: Waitlist,
     available: VecDeque<IdlingConn>,
     exist: usize,
     // only used to spawn the recycler the first time we're in async context
@@ -79,14 +90,94 @@ impl Exchange {
             // Spawn the Recycler.
             tokio::spawn(Recycler::new(pool_opts.clone(), inner.clone(), dropped));
 
-            // Spawn the ttl check interval if `inactive_connection_ttl` isn't `0`
-            if pool_opts.inactive_connection_ttl() > Duration::from_secs(0) {
+            // Spawn the ttl check interval if `inactive_connection_ttl` isn't `0` or
+            // connections have an absolute TTL.
+            if pool_opts.inactive_connection_ttl() > Duration::ZERO
+                || pool_opts.abs_conn_ttl().is_some()
+            {
                 tokio::spawn(TtlCheckInterval::new(pool_opts, inner.clone()));
             }
         }
     }
 }
 
+#[derive(Default, Debug)]
+struct Waitlist {
+    queue: KeyedPriorityQueue<QueuedWaker, QueueId>,
+}
+
+impl Waitlist {
+    fn push(&mut self, waker: Waker, queue_id: QueueId) {
+        // The documentation of Future::poll says:
+        //   Note that on multiple calls to poll, only the Waker from
+        //   the Context passed to the most recent call should be
+        //   scheduled to receive a wakeup.
+        //
+        // But the the documentation of KeyedPriorityQueue::push says:
+        //   Adds new element to queue if missing key or replace its
+        //   priority if key exists. In second case doesn’t replace key.
+        //
+        // This means we have to remove first to have the most recent
+        // waker in the queue.
+        self.remove(queue_id);
+        self.queue.push(QueuedWaker { queue_id, waker }, queue_id);
+    }
+
+    fn pop(&mut self) -> Option<Waker> {
+        match self.queue.pop() {
+            Some((qw, _)) => Some(qw.waker),
+            None => None,
+        }
+    }
+
+    fn remove(&mut self, id: QueueId) {
+        self.queue.remove(&id);
+    }
+
+    fn peek_id(&mut self) -> Option<QueueId> {
+        self.queue.peek().map(|(qw, _)| qw.queue_id)
+    }
+}
+
+const QUEUE_END_ID: QueueId = QueueId(Reverse(u64::MAX));
+
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub(crate) struct QueueId(Reverse<u64>);
+
+impl QueueId {
+    fn next() -> Self {
+        static NEXT_QUEUE_ID: atomic::AtomicU64 = atomic::AtomicU64::new(0);
+        let id = NEXT_QUEUE_ID.fetch_add(1, atomic::Ordering::SeqCst);
+        QueueId(Reverse(id))
+    }
+}
+
+#[derive(Debug)]
+struct QueuedWaker {
+    queue_id: QueueId,
+    waker: Waker,
+}
+
+impl Eq for QueuedWaker {}
+
+impl Borrow<QueueId> for QueuedWaker {
+    fn borrow(&self) -> &QueueId {
+        &self.queue_id
+    }
+}
+
+impl PartialEq for QueuedWaker {
+    fn eq(&self, other: &Self) -> bool {
+        self.queue_id == other.queue_id
+    }
+}
+
+impl Hash for QueuedWaker {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        self.queue_id.hash(state)
+    }
+}
+
 /// Connection pool data.
 #[derive(Debug)]
 pub struct Inner {
@@ -131,7 +222,7 @@ impl Pool {
                 closed: false.into(),
                 exchange: Mutex::new(Exchange {
                     available: VecDeque::with_capacity(pool_opts.constraints().max()),
-                    waiting: VecDeque::new(),
+                    waiting: Waitlist::default(),
                     exist: 0,
                     recycler: Some((rx, pool_opts)),
                 }),
@@ -148,7 +239,8 @@ impl Pool {
 
     /// Async function that resolves to `Conn`.
     pub fn get_conn(&self) -> GetConn {
-        GetConn::new(self)
+        let reset_connection = self.opts.pool_opts().reset_connection();
+        GetConn::new(self, reset_connection)
     }
 
     /// Starts a new transaction.
@@ -162,15 +254,6 @@ impl Pool {
     /// **Note:** This Future won't resolve until all active connections, taken from it,
     /// are dropped or disonnected. Also all pending and new `GetConn`'s will resolve to error.
     pub fn disconnect(self) -> DisconnectPool {
-        let was_closed = self.inner.close.swap(true, atomic::Ordering::AcqRel);
-        if !was_closed {
-            // make sure we wake up the Recycler.
-            //
-            // note the lack of an .expect() here, because the Recycler may decide that there are
-            // no connections to wait for and exit quickly!
-            let _ = self.drop.send(None).is_ok();
-        }
-
         DisconnectPool::new(self)
     }
 
@@ -178,25 +261,6 @@ impl Pool {
     fn return_conn(&mut self, conn: Conn) {
         // NOTE: we're not in async context here, so we can't block or return NotReady
         // any and all cleanup work _has_ to be done in the spawned recycler
-
-        // fast-path for when the connection is immediately ready to be reused
-        if conn.inner.stream.is_some()
-            && !conn.inner.disconnected
-            && !conn.expired()
-            && conn.inner.tx_status == TxStatus::None
-            && !conn.has_pending_result()
-            && !self.inner.close.load(atomic::Ordering::Acquire)
-        {
-            let mut exchange = self.inner.exchange.lock().unwrap();
-            if exchange.available.len() < self.opts.pool_opts().active_bound() {
-                exchange.available.push_back(conn.into());
-                if let Some(w) = exchange.waiting.pop_front() {
-                    w.wake();
-                }
-                return;
-            }
-        }
-
         self.send_to_recycler(conn);
     }
 
@@ -221,21 +285,21 @@ impl Pool {
     ///
     /// Decreases the exist counter since a broken or dropped connection should not count towards
     /// the total.
-    fn cancel_connection(&self) {
+    pub(super) fn cancel_connection(&self) {
         let mut exchange = self.inner.exchange.lock().unwrap();
         exchange.exist -= 1;
         // we just enabled the creation of a new connection!
-        if let Some(w) = exchange.waiting.pop_front() {
+        if let Some(w) = exchange.waiting.pop() {
             w.wake();
         }
     }
 
     /// Poll the pool for an available connection.
-    fn poll_new_conn(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<GetConn>> {
-        self.poll_new_conn_inner(cx)
-    }
-
-    fn poll_new_conn_inner(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<GetConn>> {
+    fn poll_new_conn(
+        &mut self,
+        cx: &mut Context<'_>,
+        queue_id: QueueId,
+    ) -> Poll<Result<GetConnInner>> {
         let mut exchange = self.inner.exchange.lock().unwrap();
 
         // NOTE: this load must happen while we hold the lock,
@@ -247,18 +311,28 @@ impl Pool {
 
         exchange.spawn_futures_if_needed(&self.inner);
 
+        // Check if we are higher priority than anything current
+        let highest = if let Some(cur) = exchange.waiting.peek_id() {
+            queue_id > cur
+        } else {
+            true
+        };
+
+        // If we are not, just queue
+        if !highest {
+            exchange.waiting.push(cx.waker().clone(), queue_id);
+            return Poll::Pending;
+        }
+
         while let Some(IdlingConn { mut conn, .. }) = exchange.available.pop_back() {
             if !conn.expired() {
-                return Poll::Ready(Ok(GetConn {
-                    pool: Some(self.clone()),
-                    inner: GetConnInner::Checking(
-                        async move {
-                            conn.stream_mut()?.check().await?;
-                            Ok(conn)
-                        }
-                        .boxed(),
-                    ),
-                }));
+                return Poll::Ready(Ok(GetConnInner::Checking(
+                    async move {
+                        conn.stream_mut()?.check().await?;
+                        Ok(conn)
+                    }
+                    .boxed(),
+                )));
             } else {
                 self.send_to_recycler(conn);
             }
@@ -270,21 +344,32 @@ impl Pool {
             // we are allowed to make a new connection, so we will!
             exchange.exist += 1;
 
-            return Poll::Ready(Ok(GetConn {
-                pool: Some(self.clone()),
-                inner: GetConnInner::Connecting(Conn::new(self.opts.clone()).boxed()),
-            }));
+            return Poll::Ready(Ok(GetConnInner::Connecting(
+                Conn::new(self.opts.clone()).boxed(),
+            )));
         }
 
-        // no go -- we have to wait
-        exchange.waiting.push_back(cx.waker().clone());
+        // Polled, but no conn available? Back into the queue.
+        exchange.waiting.push(cx.waker().clone(), queue_id);
         Poll::Pending
     }
+
+    fn unqueue(&self, queue_id: QueueId) {
+        let mut exchange = self.inner.exchange.lock().unwrap();
+        exchange.waiting.remove(queue_id);
+    }
 }
 
 impl Drop for Conn {
     fn drop(&mut self) {
+        self.inner.infile_handler = None;
+
         if std::thread::panicking() {
+            // Try to decrease the number of existing connections.
+            if let Some(pool) = self.inner.pool.take() {
+                pool.cancel_connection();
+            }
+
             return;
         }
 
@@ -299,16 +384,27 @@ impl Drop for Conn {
 #[cfg(test)]
 mod test {
     use futures_util::{
-        future::{join_all, select, select_all, try_join_all},
-        try_join, FutureExt,
+        future::{join_all, select, select_all, try_join_all, Either},
+        poll, try_join, FutureExt,
     };
-    use mysql_common::row::Row;
-    use tokio::time::sleep;
+    use tokio::time::{sleep, timeout};
+    use waker_fn::waker_fn;
 
-    use std::time::Duration;
+    use std::{
+        cmp::Reverse,
+        future::Future,
+        pin::pin,
+        sync::{Arc, OnceLock},
+        task::{Context, Poll, RawWaker, RawWakerVTable, Waker},
+        time::Duration,
+    };
 
     use crate::{
-        conn::pool::Pool, opts::PoolOpts, prelude::*, test_misc::get_opts, PoolConstraints, TxOpts,
+        conn::pool::{Pool, QueueId, Waitlist, QUEUE_END_ID},
+        opts::PoolOpts,
+        prelude::*,
+        test_misc::get_opts,
+        PoolConstraints, Row, TxOpts, Value,
     };
 
     macro_rules! conn_ex_field {
@@ -323,6 +419,61 @@ mod test {
         };
     }
 
+    fn pool_with_one_connection() -> Pool {
+        let pool_opts = PoolOpts::new().with_constraints(PoolConstraints::new(1, 1).unwrap());
+        let opts = get_opts().pool_opts(pool_opts.clone());
+        Pool::new(opts)
+    }
+
+    #[tokio::test]
+    async fn should_opt_out_of_connection_reset() -> super::Result<()> {
+        let pool_opts = PoolOpts::new().with_constraints(PoolConstraints::new(1, 1).unwrap());
+        let opts = get_opts().pool_opts(pool_opts.clone());
+
+        let pool = Pool::new(opts.clone());
+
+        let mut conn = pool.get_conn().await.unwrap();
+        assert_eq!(
+            conn.query_first::<Value, _>("SELECT @foo").await?.unwrap(),
+            Value::NULL
+        );
+        conn.query_drop("SET @foo = 'foo'").await?;
+        assert_eq!(
+            conn.query_first::<String, _>("SELECT @foo").await?.unwrap(),
+            "foo",
+        );
+        drop(conn);
+
+        conn = pool.get_conn().await.unwrap();
+        assert_eq!(
+            conn.query_first::<Value, _>("SELECT @foo").await?.unwrap(),
+            Value::NULL
+        );
+        conn.query_drop("SET @foo = 'foo'").await?;
+        conn.reset_connection(false);
+        drop(conn);
+
+        conn = pool.get_conn().await.unwrap();
+        assert_eq!(
+            conn.query_first::<String, _>("SELECT @foo").await?.unwrap(),
+            "foo",
+        );
+        drop(conn);
+        pool.disconnect().await.unwrap();
+
+        let pool = Pool::new(opts.pool_opts(pool_opts.with_reset_connection(false)));
+        conn = pool.get_conn().await.unwrap();
+        conn.query_drop("SET @foo = 'foo'").await?;
+        drop(conn);
+        conn = pool.get_conn().await.unwrap();
+        assert_eq!(
+            conn.query_first::<String, _>("SELECT @foo").await?.unwrap(),
+            "foo",
+        );
+        drop(conn);
+        pool.disconnect().await
+    }
+
     #[test]
     fn should_not_hang() -> super::Result<()> {
         pub struct Database {
@@ -354,7 +505,7 @@ mod test {
 
     #[tokio::test]
     async fn should_connect() -> super::Result<()> {
-        let pool = Pool::new(get_opts());
+        let pool = Pool::new(crate::Opts::from(get_opts()));
         pool.get_conn().await?.ping().await?;
         pool.disconnect().await?;
         Ok(())
@@ -385,12 +536,15 @@ mod test {
                 .map(|conn| conn.id())
                 .collect::<Vec<_>>();
 
+            // give some time to reset connections
+            sleep(Duration::from_millis(1000)).await;
+
             // get_conn should work if connection is available and alive
             pool.get_conn().await?;
 
             // now we'll kill connections..
             for id in ids {
-                master.query_drop(&format!("KILL {}", id)).await?;
+                master.query_drop(format!("KILL {}", id)).await?;
             }
 
             // now check, that they're still in the pool..
@@ -417,6 +571,24 @@ mod test {
         master.disconnect().await
     }
 
+    #[tokio::test]
+    async fn should_reuse_connections() -> super::Result<()> {
+        let pool = pool_with_one_connection();
+        let mut conn = pool.get_conn().await?;
+
+        let server_version = conn.server_version();
+        let connection_id = conn.id();
+
+        for _ in 0..16 {
+            drop(conn);
+            conn = pool.get_conn().await?;
+            println!("CONN connection_id={}", conn.id());
+            assert!(conn.id() == connection_id || server_version < (5, 7, 2));
+        }
+
+        Ok(())
+    }
+
     #[tokio::test]
     #[ignore]
     async fn can_handle_the_pressure() {
@@ -433,32 +605,38 @@ mod test {
             }
             drop(tx);
             // see that all the tx's eventually complete
-            while let Some(_) = rx.recv().await {}
+            while (rx.recv().await).is_some() {}
         }
         drop(pool);
     }
 
     #[tokio::test]
     async fn should_start_transaction() -> super::Result<()> {
-        let constraints = PoolConstraints::new(1, 1).unwrap();
-        let opts = get_opts().pool_opts(PoolOpts::default().with_constraints(constraints));
+        let pool = pool_with_one_connection();
 
-        let pool = Pool::new(opts);
-
-        "CREATE TEMPORARY TABLE tmp(id int)".ignore(&pool).await?;
+        "CREATE TABLE IF NOT EXISTS mysql.tmp(id int)"
+            .ignore(&pool)
+            .await?;
+        "DELETE FROM mysql.tmp".ignore(&pool).await?;
 
         let mut tx = pool.start_transaction(TxOpts::default()).await?;
-        tx.exec_batch("INSERT INTO tmp (id) VALUES (?)", vec![(1_u8,), (2_u8,)])
-            .await?;
-        tx.exec_drop("SELECT * FROM tmp", ()).await?;
+        tx.exec_batch(
+            "INSERT INTO mysql.tmp (id) VALUES (?)",
+            vec![(1_u8,), (2_u8,)],
+        )
+        .await?;
+        tx.exec_drop("SELECT * FROM mysql.tmp", ()).await?;
         drop(tx);
         let row_opt = pool
             .get_conn()
             .await?
-            .query_first("SELECT COUNT(*) FROM tmp")
+            .query_first("SELECT COUNT(*) FROM mysql.tmp")
             .await?;
         assert_eq!(row_opt, Some((0u8,)));
-        pool.get_conn().await?.query_drop("DROP TABLE tmp").await?;
+        pool.get_conn()
+            .await?
+            .query_drop("DROP TABLE mysql.tmp")
+            .await?;
         pool.disconnect().await?;
         Ok(())
     }
@@ -529,7 +707,7 @@ mod test {
             let _ = conns.pop();
 
             // then, wait for a bit to let the connection be reclaimed
-            sleep(Duration::from_millis(50)).await;
+            sleep(Duration::from_millis(500)).await;
 
             // now check that we have the expected # of connections
             // this may look a little funky, but think of it this way:
@@ -727,10 +905,7 @@ mod test {
 
     #[tokio::test]
     async fn should_ignore_non_fatal_errors_while_returning_to_a_pool() -> super::Result<()> {
-        let pool_constraints = PoolConstraints::new(1, 1).unwrap();
-        let pool_opts = PoolOpts::default().with_constraints(pool_constraints);
-
-        let pool = Pool::new(get_opts().pool_opts(pool_opts));
+        let pool = pool_with_one_connection();
         let id = pool.get_conn().await?.id();
 
         // non-fatal errors are ignored
@@ -743,6 +918,24 @@ mod test {
         Ok(())
     }
 
+    #[tokio::test]
+    async fn should_remove_waker_of_cancelled_task() {
+        let pool = pool_with_one_connection();
+        let only_conn = pool.get_conn().await.unwrap();
+
+        let join_handle = tokio::spawn(timeout(Duration::from_secs(1), pool.get_conn()));
+
+        sleep(Duration::from_secs(2)).await;
+
+        match join_handle.await.unwrap() {
+            Err(_elapsed) => (),
+            _ => panic!("unexpected Ok()"),
+        }
+        drop(only_conn);
+
+        assert_eq!(0, pool.inner.exchange.lock().unwrap().waiting.queue.len());
+    }
+
     #[tokio::test]
     async fn should_work_if_pooled_connection_operation_is_cancelled() -> super::Result<()> {
         let pool = Pool::new(get_opts());
@@ -787,6 +980,180 @@ mod test {
         Ok(())
     }
 
+    #[test]
+    fn waitlist_integrity() {
+        const DATA: *const () = &();
+        const NOOP_CLONE_FN: unsafe fn(*const ()) -> RawWaker = |_| RawWaker::new(DATA, &RW_VTABLE);
+        const NOOP_FN: unsafe fn(*const ()) = |_| {};
+        static RW_VTABLE: RawWakerVTable =
+            RawWakerVTable::new(NOOP_CLONE_FN, NOOP_FN, NOOP_FN, NOOP_FN);
+        let w = unsafe { Waker::from_raw(RawWaker::new(DATA, &RW_VTABLE)) };
+
+        let mut waitlist = Waitlist::default();
+        assert_eq!(0, waitlist.queue.len());
+
+        waitlist.push(w.clone(), QueueId(Reverse(4)));
+        waitlist.push(w.clone(), QueueId(Reverse(2)));
+        waitlist.push(w.clone(), QueueId(Reverse(8)));
+        waitlist.push(w.clone(), QUEUE_END_ID);
+        waitlist.push(w.clone(), QueueId(Reverse(10)));
+
+        waitlist.remove(QueueId(Reverse(8)));
+
+        assert_eq!(4, waitlist.queue.len());
+
+        let (_, id) = waitlist.queue.pop().unwrap();
+        assert_eq!(2, id.0 .0);
+        let (_, id) = waitlist.queue.pop().unwrap();
+        assert_eq!(4, id.0 .0);
+        let (_, id) = waitlist.queue.pop().unwrap();
+        assert_eq!(10, id.0 .0);
+        let (_, id) = waitlist.queue.pop().unwrap();
+        assert_eq!(QUEUE_END_ID, id);
+
+        assert_eq!(0, waitlist.queue.len());
+    }
+
+    #[tokio::test]
+    async fn check_absolute_connection_ttl() -> super::Result<()> {
+        let constraints = PoolConstraints::new(1, 3).unwrap();
+        let pool_opts = PoolOpts::default()
+            .with_constraints(constraints)
+            .with_inactive_connection_ttl(Duration::from_secs(99))
+            .with_ttl_check_interval(Duration::from_secs(1))
+            .with_abs_conn_ttl(Some(Duration::from_secs(2)));
+
+        let pool = Pool::new(get_opts().pool_opts(pool_opts));
+
+        let conn_ttl0 = pool.get_conn().await?;
+        sleep(Duration::from_millis(1000)).await;
+        let conn_ttl1 = pool.get_conn().await?;
+        sleep(Duration::from_millis(1000)).await;
+        let conn_ttl2 = pool.get_conn().await?;
+
+        drop(conn_ttl0);
+        drop(conn_ttl1);
+        drop(conn_ttl2);
+        assert_eq!(ex_field!(pool, exist), 3);
+
+        sleep(Duration::from_millis(1500)).await;
+        assert_eq!(ex_field!(pool, exist), 2);
+
+        sleep(Duration::from_millis(1000)).await;
+        assert_eq!(ex_field!(pool, exist), 1);
+
+        // Go even below min pool size.
+        sleep(Duration::from_millis(1000)).await;
+        assert_eq!(ex_field!(pool, exist), 0);
+
+        Ok(())
+    }
+
+    #[tokio::test]
+    async fn save_last_waker() {
+        // Test that if passed multiple wakers, we call the last one.
+
+        let pool = pool_with_one_connection();
+
+        // Get a connection, so we know the next future will be
+        // queued.
+        let conn = pool.get_conn().await.unwrap();
+        let mut pending_fut = pin!(pool.get_conn());
+
+        let build_waker = || {
+            let called = Arc::new(OnceLock::new());
+            let called2 = called.clone();
+            let waker = waker_fn(move || called2.set(()).unwrap());
+            (called, waker)
+        };
+
+        let mut assert_pending = |waker| {
+            let mut context = Context::from_waker(&waker);
+            let p = pending_fut.as_mut().poll(&mut context);
+            assert!(matches!(p, Poll::Pending));
+        };
+
+        let (first_called, waker) = build_waker();
+        assert_pending(waker);
+
+        let (second_called, waker) = build_waker();
+        assert_pending(waker);
+
+        drop(conn);
+
+        while second_called.get().is_none() {
+            assert!(first_called.get().is_none());
+            tokio::time::sleep(Duration::from_millis(100)).await;
+        }
+
+        assert!(first_called.get().is_none());
+    }
+
+    #[tokio::test]
+    async fn check_priorities() -> super::Result<()> {
+        let pool = pool_with_one_connection();
+
+        let queue_len = || {
+            let exchange = pool.inner.exchange.lock().unwrap();
+            exchange.waiting.queue.len()
+        };
+
+        // Get a connection, so we know the next futures will be
+        // queued.
+        let conn = pool.get_conn().await.unwrap();
+
+        #[allow(clippy::async_yields_async)]
+        let get_pending = || async {
+            let fut = async {
+                pool.get_conn().await.unwrap();
+            }
+            .shared();
+            let p = poll!(fut.clone());
+            assert!(matches!(p, Poll::Pending));
+            fut
+        };
+
+        let fut1 = get_pending().await;
+        let fut2 = get_pending().await;
+
+        // Both futures are queued
+        assert_eq!(queue_len(), 2);
+
+        drop(conn); // This will pop fut1 from the queue, making it [2]
+        while queue_len() != 1 {
+            tokio::time::sleep(Duration::from_millis(100)).await;
+        }
+
+        // We called wake on fut1, and even with the select fut1 will
+        // resolve first
+        let Either::Right((_, fut2)) = select(fut2, fut1).await else {
+            panic!("wrong future");
+        };
+
+        // We dropped the connection of fut1, but very likely hasn't
+        // made it through the recycler yet.
+        assert_eq!(queue_len(), 1);
+
+        let p = poll!(fut2.clone());
+        assert!(matches!(p, Poll::Pending));
+        assert_eq!(queue_len(), 1); // The queue still has fut2
+
+        // The connection will pass by the recycler and unblock fut2
+        // and pop it from the queue.
+        fut2.await;
+        assert_eq!(queue_len(), 0);
+
+        // The recycler is probably not done, so a new future will be
+        // pending.
+        let fut3 = get_pending().await;
+        assert_eq!(queue_len(), 1);
+
+        // It is OK to await it.
+        fut3.await;
+
+        Ok(())
+    }
+
     #[cfg(feature = "nightly")]
     mod bench {
         use futures_util::future::{FutureExt, TryFutureExt};
