@@ -10,11 +10,9 @@ pub use self::{read_packet::ReadPacket, write_packet::WritePacket};
 
 use bytes::BytesMut;
 use futures_core::{ready, stream};
-use futures_util::stream::{FuturesUnordered, StreamExt};
-use mio::net::{TcpKeepalive, TcpSocket};
 use mysql_common::proto::codec::PacketCodec as PacketCodecInner;
-use native_tls::{Certificate, Identity, TlsConnector};
 use pin_project::pin_project;
+use socket2::{Socket as Socket2Socket, TcpKeepalive};
 #[cfg(unix)]
 use tokio::io::AsyncWriteExt;
 use tokio::{
@@ -27,26 +25,29 @@ use tokio_util::codec::{Decoder, Encoder, Framed, FramedParts};
 use std::path::Path;
 use std::{
     fmt,
-    fs::File,
     future::Future,
     io::{
         self,
         ErrorKind::{BrokenPipe, NotConnected, Other},
-        Read,
     },
     mem::replace,
-    net::{SocketAddr, ToSocketAddrs},
     ops::{Deref, DerefMut},
     pin::Pin,
     task::{Context, Poll},
     time::Duration,
 };
 
-use crate::{buffer_pool::PooledBuf, error::IoError, opts::SslOpts};
+use crate::{
+    buffer_pool::PooledBuf,
+    error::IoError,
+    opts::{HostPortOrUrl, SslOpts, DEFAULT_PORT},
+};
 
 #[cfg(unix)]
 use crate::io::socket::Socket;
 
+mod tls;
+
 macro_rules! with_interrupted {
     ($e:expr) => {
         loop {
@@ -117,7 +118,10 @@ impl Encoder<PooledBuf> for PacketCodec {
 #[derive(Debug)]
 pub(crate) enum Endpoint {
     Plain(Option<TcpStream>),
+    #[cfg(feature = "native-tls-tls")]
     Secure(#[pin] tokio_native_tls::TlsStream<TcpStream>),
+    #[cfg(feature = "rustls-tls")]
+    Secure(#[pin] tokio_rustls::client::TlsStream<tokio::net::TcpStream>),
     #[cfg(unix)]
     Socket(#[pin] Socket),
 }
@@ -149,6 +153,11 @@ impl Future for CheckTcpStream<'_> {
 }
 
 impl Endpoint {
+    #[cfg(unix)]
+    fn is_socket(&self) -> bool {
+        matches!(self, Self::Socket(_))
+    }
+
     /// Checks, that connection is alive.
     async fn check(&mut self) -> std::result::Result<(), IoError> {
         //return Ok(());
@@ -157,89 +166,61 @@ impl Endpoint {
                 CheckTcpStream(stream).await?;
                 Ok(())
             }
+            #[cfg(feature = "native-tls-tls")]
             Endpoint::Secure(tls_stream) => {
                 CheckTcpStream(tls_stream.get_mut().get_mut().get_mut()).await?;
                 Ok(())
             }
+            #[cfg(feature = "rustls-tls")]
+            Endpoint::Secure(tls_stream) => {
+                let stream = tls_stream.get_mut().0;
+                CheckTcpStream(stream).await?;
+                Ok(())
+            }
             #[cfg(unix)]
             Endpoint::Socket(socket) => {
-                socket.write(&[]).await?;
+                let _ = socket.write(&[]).await?;
                 Ok(())
             }
             Endpoint::Plain(None) => unreachable!(),
         }
     }
 
+    #[cfg(any(feature = "native-tls-tls", feature = "rustls-tls"))]
     pub fn is_secure(&self) -> bool {
         matches!(self, Endpoint::Secure(_))
     }
 
+    #[cfg(all(not(feature = "native-tls-tls"), not(feature = "rustls")))]
+    pub async fn make_secure(
+        &mut self,
+        _domain: String,
+        _ssl_opts: crate::SslOpts,
+    ) -> crate::error::Result<()> {
+        panic!(
+            "Client had asked for TLS connection but TLS support is disabled. \
+            Please enable one of the following features: [\"native-tls-tls\", \"rustls-tls\"]"
+        )
+    }
+
     pub fn set_tcp_nodelay(&self, val: bool) -> io::Result<()> {
         match *self {
             Endpoint::Plain(Some(ref stream)) => stream.set_nodelay(val)?,
             Endpoint::Plain(None) => unreachable!(),
+            #[cfg(feature = "native-tls-tls")]
             Endpoint::Secure(ref stream) => {
                 stream.get_ref().get_ref().get_ref().set_nodelay(val)?
             }
+            #[cfg(feature = "rustls-tls")]
+            Endpoint::Secure(ref stream) => {
+                let stream = stream.get_ref().0;
+                stream.set_nodelay(val)?;
+            }
             #[cfg(unix)]
             Endpoint::Socket(_) => (/* inapplicable */),
         }
         Ok(())
     }
-
-    pub async fn make_secure(
-        &mut self,
-        domain: String,
-        ssl_opts: SslOpts,
-    ) -> std::result::Result<(), IoError> {
-        #[cfg(unix)]
-        if let Endpoint::Socket(_) = self {
-            // inapplicable
-            return Ok(());
-        }
-
-        let mut builder = TlsConnector::builder();
-        if let Some(root_cert_path) = ssl_opts.root_cert_path() {
-            let mut root_cert_data = vec![];
-            let mut root_cert_file = File::open(root_cert_path)?;
-            root_cert_file.read_to_end(&mut root_cert_data)?;
-
-            let root_certs = Certificate::from_der(&*root_cert_data)
-                .map(|x| vec![x])
-                .or_else(|_| {
-                    pem::parse_many(&*root_cert_data)
-                        .iter()
-                        .map(pem::encode)
-                        .map(|s| Certificate::from_pem(s.as_bytes()))
-                        .collect()
-                })?;
-
-            for root_cert in root_certs {
-                builder.add_root_certificate(root_cert);
-            }
-        }
-        if let Some(pkcs12_path) = ssl_opts.pkcs12_path() {
-            let der = std::fs::read(pkcs12_path)?;
-            let identity = Identity::from_pkcs12(&*der, ssl_opts.password().unwrap_or(""))?;
-            builder.identity(identity);
-        }
-        builder.danger_accept_invalid_hostnames(ssl_opts.skip_domain_validation());
-        builder.danger_accept_invalid_certs(ssl_opts.accept_invalid_certs());
-        let tls_connector: tokio_native_tls::TlsConnector = builder.build()?.into();
-
-        *self = match self {
-            Endpoint::Plain(stream) => {
-                let stream = stream.take().unwrap();
-                let tls_stream = tls_connector.connect(&*domain, stream).await?;
-                Endpoint::Secure(tls_stream)
-            }
-            Endpoint::Secure(_) => unreachable!(),
-            #[cfg(unix)]
-            Endpoint::Socket(_) => unreachable!(),
-        };
-
-        Ok(())
-    }
 }
 
 impl From<TcpStream> for Endpoint {
@@ -255,12 +236,17 @@ impl From<Socket> for Endpoint {
     }
 }
 
+#[cfg(feature = "native-tls-tls")]
 impl From<tokio_native_tls::TlsStream<TcpStream>> for Endpoint {
     fn from(stream: tokio_native_tls::TlsStream<TcpStream>) -> Self {
         Endpoint::Secure(stream)
     }
 }
 
+/* TODO
+#[cfg(feature = "rustls-tls")]
+*/
+
 impl AsyncRead for Endpoint {
     fn poll_read(
         self: Pin<&mut Self>,
@@ -272,6 +258,9 @@ impl AsyncRead for Endpoint {
             EndpointProj::Plain(ref mut stream) => {
                 Pin::new(stream.as_mut().unwrap()).poll_read(cx, buf)
             }
+            #[cfg(feature = "native-tls-tls")]
+            EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_read(cx, buf),
+            #[cfg(feature = "rustls-tls")]
             EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_read(cx, buf),
             #[cfg(unix)]
             EndpointProj::Socket(ref mut stream) => stream.as_mut().poll_read(cx, buf),
@@ -290,6 +279,9 @@ impl AsyncWrite for Endpoint {
             EndpointProj::Plain(ref mut stream) => {
                 Pin::new(stream.as_mut().unwrap()).poll_write(cx, buf)
             }
+            #[cfg(feature = "native-tls-tls")]
+            EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_write(cx, buf),
+            #[cfg(feature = "rustls-tls")]
             EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_write(cx, buf),
             #[cfg(unix)]
             EndpointProj::Socket(ref mut stream) => stream.as_mut().poll_write(cx, buf),
@@ -305,6 +297,9 @@ impl AsyncWrite for Endpoint {
             EndpointProj::Plain(ref mut stream) => {
                 Pin::new(stream.as_mut().unwrap()).poll_flush(cx)
             }
+            #[cfg(feature = "native-tls-tls")]
+            EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_flush(cx),
+            #[cfg(feature = "rustls-tls")]
             EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_flush(cx),
             #[cfg(unix)]
             EndpointProj::Socket(ref mut stream) => stream.as_mut().poll_flush(cx),
@@ -320,6 +315,9 @@ impl AsyncWrite for Endpoint {
             EndpointProj::Plain(ref mut stream) => {
                 Pin::new(stream.as_mut().unwrap()).poll_shutdown(cx)
             }
+            #[cfg(feature = "native-tls-tls")]
+            EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_shutdown(cx),
+            #[cfg(feature = "rustls-tls")]
             EndpointProj::Secure(ref mut stream) => stream.as_mut().poll_shutdown(cx),
             #[cfg(unix)]
             EndpointProj::Socket(ref mut stream) => stream.as_mut().poll_shutdown(cx),
@@ -354,108 +352,41 @@ impl Stream {
         }
     }
 
-    pub(crate) async fn connect_tcp<S>(addr: S, keepalive: Option<Duration>) -> io::Result<Stream>
-    where
-        S: ToSocketAddrs,
-    {
-        // TODO: Use tokio to setup keepalive (see tokio-rs/tokio#3082)
-        async fn connect_stream(
-            addr: SocketAddr,
-            keepalive_opts: Option<TcpKeepalive>,
-        ) -> io::Result<TcpStream> {
-            let socket = if addr.is_ipv6() {
-                TcpSocket::new_v6()?
-            } else {
-                TcpSocket::new_v4()?
-            };
-
-            if let Some(keepalive_opts) = keepalive_opts {
-                socket.set_keepalive_params(keepalive_opts)?;
+    pub(crate) async fn connect_tcp(
+        addr: &HostPortOrUrl,
+        keepalive: Option<Duration>,
+    ) -> io::Result<Stream> {
+        let tcp_stream = match addr {
+            HostPortOrUrl::HostPort(host, port) => {
+                TcpStream::connect((host.as_str(), *port)).await?
             }
+            HostPortOrUrl::Url(url) => {
+                let addrs = url.socket_addrs(|| Some(DEFAULT_PORT))?;
+                TcpStream::connect(&*addrs).await?
+            }
+        };
 
-            let stream = tokio::task::spawn_blocking(move || {
-                let mut stream = socket.connect(addr)?;
-                let mut poll = mio::Poll::new()?;
-                let mut events = mio::Events::with_capacity(1024);
-
-                poll.registry()
-                    .register(&mut stream, mio::Token(0), mio::Interest::WRITABLE)?;
-
-                loop {
-                    poll.poll(&mut events, None)?;
-
-                    for event in &events {
-                        if event.token() == mio::Token(0) && event.is_error() {
-                            return Err(io::Error::new(
-                                io::ErrorKind::ConnectionRefused,
-                                "Connection refused",
-                            ));
-                        }
-
-                        if event.token() == mio::Token(0) && event.is_writable() {
-                            // The socket connected (probably, it could still be a spurious
-                            // wakeup)
-                            return Ok::<_, io::Error>(stream);
-                        }
-                    }
-                }
-            })
-            .await??;
-
+        if let Some(duration) = keepalive {
             #[cfg(unix)]
-            let std_stream = unsafe {
+            let socket = {
                 use std::os::unix::prelude::*;
-                let fd = stream.into_raw_fd();
-                std::net::TcpStream::from_raw_fd(fd)
+                let fd = tcp_stream.as_raw_fd();
+                unsafe { Socket2Socket::from_raw_fd(fd) }
             };
-
             #[cfg(windows)]
-            let std_stream = unsafe {
+            let socket = {
                 use std::os::windows::prelude::*;
-                let fd = stream.into_raw_socket();
-                std::net::TcpStream::from_raw_socket(fd)
+                let sock = tcp_stream.as_raw_socket();
+                unsafe { Socket2Socket::from_raw_socket(sock) }
             };
-
-            Ok(TcpStream::from_std(std_stream)?)
+            socket.set_tcp_keepalive(&TcpKeepalive::new().with_time(duration))?;
+            std::mem::forget(socket);
         }
 
-        let keepalive_opts = keepalive.map(|time| TcpKeepalive::new().with_time(time));
-
-        match addr.to_socket_addrs() {
-            Ok(addresses) => {
-                let mut streams = FuturesUnordered::new();
-
-                for address in addresses {
-                    streams.push(connect_stream(address, keepalive_opts.clone()));
-                }
-
-                let mut err = None;
-                while let Some(stream) = streams.next().await {
-                    match stream {
-                        Err(e) => {
-                            err = Some(e);
-                        }
-                        Ok(stream) => {
-                            return Ok(Stream {
-                                closed: false,
-                                codec: Box::new(Framed::new(stream.into(), PacketCodec::default()))
-                                    .into(),
-                            });
-                        }
-                    }
-                }
-
-                if let Some(e) = err {
-                    Err(e)
-                } else {
-                    Err(io::Error::new(
-                        io::ErrorKind::InvalidInput,
-                        "could not resolve to any address",
-                    ))
-                }
-            }
-            Err(err) => Err(err),
-        }
+        Ok(Stream {
+            closed: false,
+            codec: Box::new(Framed::new(tcp_stream.into(), PacketCodec::default())).into(),
+        })
     }
 
     #[cfg(unix)]
@@ -480,10 +411,16 @@ impl Stream {
         Ok(())
     }
 
+    #[cfg(any(feature = "native-tls-tls", feature = "rustls-tls"))]
     pub(crate) fn is_secure(&self) -> bool {
         self.codec.as_ref().unwrap().get_ref().is_secure()
     }
 
+    #[cfg(unix)]
+    pub(crate) fn is_socket(&self) -> bool {
+        self.codec.as_ref().unwrap().get_ref().is_socket()
+    }
+
     pub(crate) fn reset_seq_id(&mut self) {
         if let Some(codec) = self.codec.as_mut() {
             codec.codec_mut().reset_seq_id();
@@ -560,6 +497,9 @@ mod test {
         let endpoint = stream.codec.as_mut().unwrap().get_ref();
         let stream = match endpoint {
             super::Endpoint::Plain(Some(stream)) => stream,
+            #[cfg(feature = "rustls-tls")]
+            super::Endpoint::Secure(tls_stream) => tls_stream.get_ref().0,
+            #[cfg(feature = "native-tls-tls")]
             super::Endpoint::Secure(tls_stream) => tls_stream.get_ref().get_ref().get_ref(),
             _ => unreachable!(),
         };
