@@ -8,8 +8,8 @@
 
 use futures_util::FutureExt;
 use mysql_common::{
+    constants::MAX_PAYLOAD_LEN,
     io::ParseBuf,
-    packets::{OkPacketDeserializer, ResultSetTerminator},
     proto::{Binary, Text},
     row::RowDeserializer,
     value::ServerSide,
@@ -28,7 +28,9 @@ use crate::{
     consts::CapabilityFlags,
     error::*,
     prelude::{FromRow, StatementLike},
+    query::AsQuery,
     queryable::query_result::ResultSetMeta,
+    tracing_utils::{LevelInfo, LevelTrace, TracingLevel},
     BoxFuture, Column, Conn, Params, ResultSetStream, Row,
 };
 
@@ -41,10 +43,11 @@ pub trait Protocol: fmt::Debug + Send + Sync + 'static {
     fn result_set_meta(columns: Arc<[Column]>) -> ResultSetMeta;
     fn read_result_set_row(packet: &[u8], columns: Arc<[Column]>) -> Result<Row>;
     fn is_last_result_set_packet(capabilities: CapabilityFlags, packet: &[u8]) -> bool {
-        packet.len() < 8
-            && ParseBuf(packet)
-                .parse::<OkPacketDeserializer<ResultSetTerminator>>(capabilities)
-                .is_ok()
+        if capabilities.contains(CapabilityFlags::CLIENT_DEPRECATE_EOF) {
+            packet[0] == 0xFE && packet.len() < MAX_PAYLOAD_LEN
+        } else {
+            packet[0] == 0xFE && packet.len() < 8
+        }
     }
 }
 
@@ -100,13 +103,32 @@ impl Conn {
     }
 
     /// Low level function that performs a text query.
-    pub(crate) async fn raw_query<'a, Q>(&'a mut self, query: Q) -> Result<()>
+    pub(crate) async fn raw_query<'a, Q, L: TracingLevel>(&'a mut self, query: Q) -> Result<()>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
     {
-        self.routine(QueryRoutine::new(query.as_ref().as_bytes()))
+        self.routine(QueryRoutine::<'_, L>::new(query.as_query().as_ref()))
             .await
     }
+
+    /// Used for internal querying of connection settings,
+    /// bypassing instrumentation meant for user queries.
+    // This is a merge of `Queryable::query_first` and `Conn::query_iter`.
+    // TODO: find a cleaner way without duplicating code.
+    pub(crate) fn query_internal<'a, T, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Option<T>>
+    where
+        Q: AsQuery + 'a,
+        T: FromRow + Send + 'static,
+    {
+        async move {
+            self.raw_query::<'_, _, LevelTrace>(query).await?;
+            Ok(QueryResult::<'_, '_, TextProtocol>::new(self)
+                .collect_and_drop::<T>()
+                .await?
+                .pop())
+        }
+        .boxed()
+    }
 }
 
 /// Methods of this trait are used to execute database queries.
@@ -122,7 +144,7 @@ pub trait Queryable: Send {
         query: Q,
     ) -> BoxFuture<'a, QueryResult<'a, 'static, TextProtocol>>
     where
-        Q: AsRef<str> + Send + Sync + 'a;
+        Q: AsQuery + 'a;
 
     /// Prepares the given statement.
     ///
@@ -134,7 +156,7 @@ pub trait Queryable: Send {
     /// [stmt_cache_size]: crate::Opts::stmt_cache_size
     fn prep<'a, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Statement>
     where
-        Q: AsRef<str> + Sync + Send + 'a;
+        Q: AsQuery + 'a;
 
     /// Closes the given statement.
     ///
@@ -165,7 +187,7 @@ pub trait Queryable: Send {
     /// to make this conversion infallible.
     fn query<'a, T, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Vec<T>>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
         T: FromRow + Send + 'static,
     {
         async move { self.query_iter(query).await?.collect_and_drop::<T>().await }.boxed()
@@ -180,7 +202,7 @@ pub trait Queryable: Send {
     /// to make this conversion infallible.
     fn query_first<'a, T, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Option<T>>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
         T: FromRow + Send + 'static,
     {
         async move {
@@ -205,7 +227,7 @@ pub trait Queryable: Send {
     /// to make this conversion infallible.
     fn query_map<'a, T, F, Q, U>(&'a mut self, query: Q, mut f: F) -> BoxFuture<'a, Vec<U>>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
         T: FromRow + Send + 'static,
         F: FnMut(T) -> U + Send + 'a,
         U: Send,
@@ -229,7 +251,7 @@ pub trait Queryable: Send {
     /// to make this conversion infallible.
     fn query_fold<'a, T, F, Q, U>(&'a mut self, query: Q, init: U, mut f: F) -> BoxFuture<'a, U>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
         T: FromRow + Send + 'static,
         F: FnMut(U, T) -> U + Send + 'a,
         U: Send + 'a,
@@ -246,12 +268,12 @@ pub trait Queryable: Send {
     /// Performs the given query and drops the query result.
     fn query_drop<'a, Q>(&'a mut self, query: Q) -> BoxFuture<'a, ()>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
     {
         async move { self.query_iter(query).await?.drop_result().await }.boxed()
     }
 
-    /// Exectues the given statement for each item in the given params iterator.
+    /// Executes the given statement for each item in the given params iterator.
     ///
     /// It'll prepare `stmt` (once), if necessary.
     fn exec_batch<'a: 'b, 'b, S, P, I>(&'a mut self, stmt: S, params_iter: I) -> BoxFuture<'b, ()>
@@ -261,7 +283,7 @@ pub trait Queryable: Send {
         I::IntoIter: Send,
         P: Into<Params> + Send;
 
-    /// Exectues the given statement and collects the first result set.
+    /// Executes the given statement and collects the first result set.
     ///
     /// It'll prepare `stmt`, if necessary.
     ///
@@ -285,7 +307,7 @@ pub trait Queryable: Send {
         .boxed()
     }
 
-    /// Exectues the given statement and returns the first row of the first result set.
+    /// Executes the given statement and returns the first row of the first result set.
     ///
     /// It'll prepare `stmt`, if necessary.
     ///
@@ -313,7 +335,7 @@ pub trait Queryable: Send {
         .boxed()
     }
 
-    /// Exectues the given stmt and maps each row of the first result set.
+    /// Executes the given stmt and maps each row of the first result set.
     ///
     /// It'll prepare `stmt`, if necessary.
     ///
@@ -345,7 +367,7 @@ pub trait Queryable: Send {
         .boxed()
     }
 
-    /// Exectues the given stmt and folds the first result set to a signel value.
+    /// Executes the given stmt and folds the first result set to a signel value.
     ///
     /// It'll prepare `stmt`, if necessary.
     ///
@@ -377,7 +399,7 @@ pub trait Queryable: Send {
         .boxed()
     }
 
-    /// Exectues the given statement and drops the result.
+    /// Executes the given statement and drops the result.
     fn exec_drop<'a: 'b, 'b, S, P>(&'a mut self, stmt: S, params: P) -> BoxFuture<'b, ()>
     where
         S: StatementLike + 'b,
@@ -397,7 +419,7 @@ pub trait Queryable: Send {
     ) -> BoxFuture<'a, ResultSetStream<'a, 'a, 'static, T, TextProtocol>>
     where
         T: Unpin + FromRow + Send + 'static,
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
     {
         async move {
             self.query_iter(query)
@@ -451,11 +473,10 @@ impl Queryable for Conn {
         query: Q,
     ) -> BoxFuture<'a, QueryResult<'a, 'static, TextProtocol>>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
     {
         async move {
-            self.routine(QueryRoutine::new(query.as_ref().as_bytes()))
-                .await?;
+            self.raw_query::<'_, _, LevelInfo>(query).await?;
             Ok(QueryResult::new(self))
         }
         .boxed()
@@ -463,9 +484,9 @@ impl Queryable for Conn {
 
     fn prep<'a, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Statement>
     where
-        Q: AsRef<str> + Sync + Send + 'a,
+        Q: AsQuery + 'a,
     {
-        async move { self.get_statement(query.as_ref()).await }.boxed()
+        async move { self.get_statement(query.as_query()).await }.boxed()
     }
 
     fn close(&mut self, stmt: Statement) -> BoxFuture<'_, ()> {
@@ -525,14 +546,14 @@ impl Queryable for Transaction<'_> {
         query: Q,
     ) -> BoxFuture<'a, QueryResult<'a, 'static, TextProtocol>>
     where
-        Q: AsRef<str> + Send + Sync + 'a,
+        Q: AsQuery + 'a,
     {
         self.0.query_iter(query)
     }
 
     fn prep<'a, Q>(&'a mut self, query: Q) -> BoxFuture<'a, Statement>
     where
-        Q: AsRef<str> + Sync + Send + 'a,
+        Q: AsQuery + 'a,
     {
         self.0.prep(query)
     }
@@ -566,7 +587,6 @@ impl Queryable for Transaction<'_> {
 
 #[cfg(test)]
 mod tests {
-    use super::Queryable;
     use crate::{error::Result, prelude::*, test_misc::get_opts, Conn};
 
     #[tokio::test]
